/* Processing code.

   OperStats - Copyright (c) 2000-2002 Darcy Grexton
   Contact: skold@habber.net, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#include "../inc/operstats.h"

Servers *servlist = NULL;
ServStat *statlist[1024];
int serv_size = 0;

/* split_buf:  Split a buffer into arguments and store the arguments in an
               argument vector pointed to by argv (which will be malloc'd
               as necessary); return the argument count.  If colon_special
               is non-zero, then treat a parameter with a leading ':' as
               the last parameter of the line, per the IRC RFC. Destroys
               the buffer by side effect.
 */
int split_buf (char *buf, char ***argv, int colon_special)
{
    int argvsize = 8;
    int argc;
    char *s;

    *argv = smalloc (sizeof(char *) * argvsize);
    argc = 0;
    while (*buf)
    {
	if (argc == argvsize)
	{
	    argvsize += 8;
	    *argv = srealloc (*argv, sizeof(char *) * argvsize);
	}
	if (*buf == ':')
	{
	    (*argv)[argc++] = buf+1;
	    buf = "";
	}
	else
	{
	    s = strpbrk(buf, " ");
	    if (s)
	    {
		*s++ = 0;
		while (isspace(*s))
		    s++;
	    }
	    else
		s = buf + strlen(buf);
	    (*argv)[argc++] = buf;
	    buf = s;
	}
    }
    return argc;
}

/* process:  Main processing routine. */
void process ()
{
    char source[64];
    char cmd[64];
    char buf[512];
    char *s;
    int ac;
    char **av;
    Message *m;

    /* If debugging, log the buffer. */
    if (debuglevel == 1 || debuglevel == 3)
	debug ("--> %s", inbuf);

    /* First make a copy of the buffer so we have the original in case we
       crash - in that case, we want to know what we crashed on.
     */
    strscpy (buf, inbuf, sizeof (buf));

    /* Split the buffer into pieces. */
    if (*buf == ':')
    {
	s = strpbrk (buf, " ");
	if (!s)
	    return;
	*s = 0;
	while (isspace (*s++))
	    ;
	strscpy (source, buf+1, sizeof(source));
	memmove (buf, s, strlen (s)+1);
    }
    else
	*source = 0;
    if (!*buf)
	return;
    s = strpbrk (buf, " ");
    if (s)
    {
	*s = 0;
	while (isspace (*s++))
	    ;
    }
    else
	s = buf + strlen(buf);
    strscpy (cmd, buf, sizeof(cmd));
    ac = split_buf (s, &av, 1);

    /* Do something with the message. */
    m = find_message (cmd);

    if (m)
	if (m->func)
	    m->func (source, ac, av);

    /* Free argument list we created */
    free (av);
}

/* Send our MOTD */
static void m_motd (char *source, int ac, char **av)
{
    FILE *f;
    char buf[BUFSIZE];

    send_cmd (me.name, "375 %s :- %s Message of the Day -", source, me.name);
    send_cmd (me.name, "372 %s :- ", source);
    send_cmd (me.name,
	"372 %s :- OperStats v%s Copyright (c) 2000-2002 Darcy Grexton",
	source, version);
    send_cmd (me.name, "372 %s :- ", source);

    f = fopen (MOTDFILE, "r");

    if (f)
    {
	while (fgets (buf, sizeof (buf), f))
	{
	    buf[strlen (buf) - 1] = 0;
	    send_cmd (me.name, "372 %s :- %s", source, buf);
	}
	fclose (f);
    }

    send_cmd (me.name, "376 %s :End of /MOTD command.", source);
}

/* Reply to a TIME query */
static void m_time (char *source, int ac, char **av)
{
    time_t t;
    struct tm *tm;
    char buf[80];

    time (&t);
    tm = localtime (&t);
    strftime (buf, sizeof (buf), "%A %B %d %Y -- %H:%M %z", tm);

    send_cmd (me.name, "391 %s %s :%s", source, me.name, buf);
}

/* Send an ADMIN reply */
static void m_admin (char *source, int ac, char **av)
{
    send_cmd (me.name, "256 %s :Administrative info about %s", source,
	me.name);
    send_cmd (me.name, "257 %s :OperStats v%s", source, version);
    send_cmd (me.name, "258 %s :Copyright (c) 2000-2002 Darcy Grexton",
	source);
    send_cmd (me.name,
	"259 %s :For more information E-Mail skold@habber.net.", source);
}

/* Reply to a ping */
static void m_ping (char *source, int ac, char **av)
{
    if (ac < 1)
	return;

    send_cmd (me.name, "%s %s %s", me.token ? "9" : "PONG",
	ac > 1 ? av[1] : me.name, av[0]);
}

/* A server replied to a PING, update its ping time */
static void m_pong (char *source, int ac, char **av)
{
    ServStat *serverstat;
    int i;

    serverstat = findserv (source);

    if (ac < 1)
	return;

    for (i = 0; i < servcnt && stricmp (servlist[i].server, source); i++);
	if ((i < servcnt) && serverstat)
	{
	    struct timeval now, tmp;
	    gettimeofday (&now, NULL);
	    timersub (&now, &(servlist[i].lastping), &tmp);

	    serverstat->lag.tv_sec = tmp.tv_sec;
	    serverstat->lag.tv_usec = tmp.tv_usec;
	    servlist[i].lastpong.tv_sec = now.tv_sec;
	    servlist[i].lastpong.tv_usec = now.tv_usec;

	    if (timercmp (&tmp, &(serverstat->peaklag), >))
	    {
		serverstat->peaklag.tv_sec = tmp.tv_sec;
		serverstat->peaklag.tv_usec = tmp.tv_usec;
	    }

	    if (globalonlag_on == TRUE)
	    {
		/* If theres some pocket lag involved, we may get several
		   PONGs at once. To avoid flooding in globops about it,
		   we only report lag once every ping_delay seconds.
		 */
		if (tv2ms (&tmp) >= (lagthreshold * 1000))
		{
		    if (now.tv_sec - serverstat->lastreport < ping_delay)
			return;
		    else
		    {
			globops (s_StatServ, SS_SERVER_LAGGED, source, tv2ms (&tmp));
			serverstat->lastreport = now.tv_sec;
		    }
		}
	    }
	}
}

/* Handle a KICK command */
static void m_kick (char *source, int ac, char **av)
{
    if (ac != 3)
	return; 

    do_kick (source, ac, av);
}

/* Handle a TOPIC command */
static void m_topic (char *source, int ac, char **av)
{
    if (ac != 4)
	return;

    do_topic (source, ac, av);
}

/* Handle a JOIN command */
static void m_join (char *source, int ac, char **av)
{
    if (ac != 1)
	return;

    do_join (source, ac, av);
}

#if defined (BAHAMUT) || defined (PROMETHEUS)
/* Handle an SJOIN command. */
static void m_sjoin (char *source, int ac, char **av)
{
    do_sjoin (source, ac, av);
}
#endif

/* Handle a PART command */
static void m_part (char *source, int ac, char **av)
{
    if (ac < 1 || ac > 2)   
	return;

    do_part (source, ac, av);
}

/* An AKILL was placed elsewhere, learn it. */
static void m_akill (char *source, int ac, char **av)
{
    AutoKill *akill, **list;
    char *mask;

    /* Only do this if OperServ is on. */
    if (operserv_on == FALSE)
	return;

    /* Should we learn new AKills? */
    if (learnakills_on == FALSE)
	return;

    mask = smalloc (strlen (av[1]) + strlen (av[0]) + 2);
    snprintf (mask, strlen (av[1]) + strlen (av[0]) + 2, "%s@%s", av[1],
	av[0]);

    /* Ignore it if we already have it. */
    if (is_akilled (mask))
	return;

    akill = scalloc (sizeof (AutoKill), 1);
    akill->mask = sstrdup (mask);
    akill->reason = sstrdup (av[2]);
    akill->setter = sstrdup (source);
    akill->realname = 0;
    akill->set = time (NULL);
    akill->expires = def_akill_time;
    list = &akilllist[AHASH(akill->mask)];
    akill->next = *list;
    if (*list)
	(*list)->prev = akill;
    *list = akill;

    if (globalonakill_on == TRUE)
	globops (s_OperServ, OS_AKILL_LEARNED, source, mask, av[2]);

    log (OS_AKILL_LEARNED, source, mask, av[2]);

    free (mask);
}

/* An AKill was removed elsewhere, remove it if we have it. */
static void m_rakill (char *source, int ac, char **av)
{
    AutoKill *akill;
    int i, j;
    char *mask;

    /* Only do this if OperServ is on. */
    if (operserv_on == FALSE)
	return;

    /* Should we forget AKills? */
    if (forgetakills_on == FALSE)
	return;

    mask = smalloc (strlen (av[1]) + strlen (av[0]) + 2);
    snprintf (mask, strlen (av[1]) + strlen (av[0]) + 2, "%s@%s", av[1],
	av[0]);

    /* Make sure we have it */
    if (!is_akilled (mask))
	return;

    for (i = 0; i < 1024; i++)
    {
	for (akill = akilllist[i]; akill; akill = akill->next)
	{
	    j++;

	    if (!stricmp (mask, akill->mask))
	    {
		free (akill->setter);
		free (akill->reason);
		if (akill->prev)
		    akill->prev->next = akill->next;
		else
		    akilllist[AHASH(akill->mask)] = akill->next;
		if (akill->next)
		    akill->next->prev = akill->prev;

		free (akill);

		if (globalonakill_on == TRUE)
		    globops (s_OperServ, OS_AKILL_FORGOTTEN, mask, source);
		log (OS_AKILL_FORGOTTEN, mask, source);

		return;
	    }
	}
    }
}

/* Handle a NICK command */
static void m_nick (char *source, int ac, char **av)
{
#ifdef IRCUP9
    if (strchr (source, '.'))
	*source = 0;
#endif

    do_nick (source, ac, av);
}

/* Handle a QUIT command */
static void m_quit (char *source, int ac, char **av)
{
    if (ac != 1)
	return;

    if (!finduser (source))
    {
	debug ("do_quit(): Got QUIT for nonexistant user %s", source);
	return;
    }

    do_quit (source, ac, av);
}

/* Send our version reply */
void m_version (char *source, int ac, char **av)
{
    if (source)
	send_cmd (me.name, "351 %s operstats-%s. %s :%s%s%s%s%s",
	    source, version, me.name,
	    (operserv_on == TRUE) ? "O" : "",
	    (statserv_on == TRUE) ? "S" : "",
	    (globalnoticer_on == TRUE) ? "g" : "",
	    (spamserv_on == TRUE) ? "s" : "",
	    debuglevel ? "d" : "");
}

/* A server replied to a version query, store its reply. */
void m_vreply (char *source, int ac, char **av)
{
    int i;
    ServStat *server;

    for (i = 0; i < 1024; i++)
    {       
	for (server = statlist[i]; server; server = server->next)
	{
	    if (!stricmp (server->name, source))
	    {
		/* Some jackasses edit their IRCd's version reply,
		   or run OperStats with unsupported servers. We need
		   to make sure av[1] and av[3] actually EXIST here,
		   otherwise we crash. If we DONT find them, they'll
		   stay as the defaults of 'Unknown' and 'XX'.
		 */
		if (av[1])
		    server->version = sstrdup (av[1]);
		if (av[3])
		    server->flags = sstrdup (av[3]);
	    }
	}
    }
}

/* Reply to an INFO command */
void m_info (char *source, int ac, char **av)
{
    int i;

    for (i = 0; infotext[i]; i++)
	send_cmd (me.name, "371 %s :%s", source, infotext[i]);
    send_cmd (me.name, "371 %s :operstats-%s", source, version);
    send_cmd (me.name,
	      "371 %s :Birth Date: %s, compile # %s", source, compile_time,
	      build);
    send_cmd (me.name, "371 %s :On-line since %s", source, ctime (&me.since));
    send_cmd (me.name, "374 %s :End of /INFO list.", source);
}

#ifdef UNREAL3

/* Reply to a NETINFO command. */
void m_netinfo (char *source, int ac, char **av)
{
    /* If the server says it's seen a higher max than we have, update
       our numbers. The time will be off, but ah well :)
     */
    if (atoi (av[0]) > maxusercnt)
    {
	maxusercnt = atoi (av[0]);
	maxusertime = time (NULL);
    }

    send_cmd (me.name, "%s %d %lu 0 * * * * :%s", me.token ? "AO" : "NETINFO",
	maxusercnt, time (NULL), av[7]);

    globops (operserv_on == TRUE ? s_OperServ : s_StatServ,
	"Sync'd to network in %s: %d users, %d channels, %d servers.",
	duration (time (NULL) - burststart, 1), usercnt, chancnt, servcnt);
}
#endif /* UNREAL3 */

/* Reply to a STATS query */
void m_stats (char *source, int ac, char **av)
{
    if (ac < 1)
	return;

    switch (*av[0])
    {
	case 'u':
	{
	    int uptime = time (NULL) - me.since;
	    int tmp = servcount ();

	    send_cmd (me.name, "242 %s :Server Up %d days, %d:%02d:%02d",
		source, uptime/86400, (uptime/3600) % 24, (uptime/60) % 60,
		uptime % 60);
	    send_cmd (me.name,
		"250 %s :Highest connection count: %d (%d clients)",
		source, tmp + 1, tmp);
	    send_cmd (me.name, "219 %s u :End of /STATS report.", source);
	    break;
	}
	default:
	    send_cmd (me.name, "219 %s * :End of /STATS report.", source);
    }
}

/* Reply to a WHOIS query */
void m_whois (char *source, int ac, char **av)
{   
    if (ac < 1)
	return;   

    if ((!stricmp (av[0], s_OperServ)) && operserv_on == TRUE)
    {
	send_cmd (me.name, "311 %s %s %s %s * :%s", source, s_OperServ,
	    os.user, os.host, os.real);
	send_cmd (me.name, "312 %s %s %s :%s", source, s_OperServ,
	    me.name, me.desc);
	if (strchr (os.mode, 'o'))
	    send_cmd (me.name, "313 %s %s :is an IRC Operator", source,
		s_OperServ);
	send_cmd (me.name, "317 %s %s 0 %d :seconds idle, signon time",
	    source, s_OperServ, me.since);
    }
    else if ((!stricmp (av[0], s_StatServ)) && statserv_on == TRUE)
    {
	send_cmd (me.name, "311 %s %s %s %s * :%s", source, s_StatServ,
	    ss.user, ss.host, ss.real);
	send_cmd (me.name, "312 %s %s %s :%s", source, s_StatServ,
	    me.name, me.desc);
	if (strchr (ss.mode, 'o'))
	    send_cmd (me.name, "313 %s %s :is an IRC Operator", source,
		s_StatServ);
	send_cmd (me.name, "317 %s %s 0 %d :seconds idle, signon time",
	    source, s_StatServ, me.since);
    }
    else if ((!stricmp (av[0], s_GlobalNoticer)) && globalnoticer_on == TRUE)
    {
	send_cmd (me.name, "311 %s %s %s %s * :%s", source, s_GlobalNoticer,
	    gn.user, gn.host, gn.real);
	send_cmd (me.name, "312 %s %s %s :%s", source, s_GlobalNoticer,
	    me.name, me.desc);
	if (strchr (gn.mode, 'o'))
	    send_cmd (me.name, "313 %s %s :is an IRC Operator", source,
		s_GlobalNoticer);
	send_cmd (me.name, "317 %s %s 0 %d :seconds idle, signon time",
	    source, s_GlobalNoticer, me.since);
    }
    else if ((!stricmp (av[0], sp.nick)) && spamserv_on == TRUE)
    {
	send_cmd (me.name, "311 %s %s %s %s * :%s", source, sp.nick,
	    sp.user, sp.host, sp.real);
	send_cmd (me.name, "312 %s %s %s :%s", source, sp.nick,
	    me.name, me.desc);
	if (strchr (ss.mode, 'o'))
	    send_cmd (me.name, "313 %s %s :is an IRC Operator", source,
		sp.nick);
	send_cmd (me.name, "317 %s %s 0 %d :seconds idle, signon time",
	    source, sp.nick, me.since);
    }
    else
	send_cmd (me.name, "401 %s %s :No such nick/channel", source, av[0]);
    send_cmd (me.name, "318 %s %s :End of /WHOIS list", source, av[0]);
}

/* Handle a SERVER command */
static void m_server (char *source, int ac, char **av)
{
    ServStat *serverstat, **list;

    if (servcnt >= serv_size)
    {
	if (serv_size < 8)
	    serv_size = 8;
	else
	    serv_size *= 2;
	servlist = srealloc (servlist, sizeof (*servlist) * serv_size);
    }
    servlist[servcnt].server = sstrdup (av[0]);
    servlist[servcnt].uplink = sstrdup (source);
    servlist[servcnt].hops = atoi (av[1]);
    servlist[servcnt].connect = time (NULL);
    gettimeofday (&(servlist[servcnt].lastping), NULL);

    /* We'll only ping if StatServ is on, because the results aren't
       viewable otherwise. The VERSION reply isn't viewable either, but
       StatServ may be enabled later on.
     */
    send_cmd (me.name, "%s %s", me.token ? "+" : "VERSION", av[0]);
    if (statserv_on == TRUE)
	send_cmd (me.name, "%s %s %s", me.token ? "8" : "PING", me.name, av[0]);
    servcnt++;
    servstat++;
    if (servstat > maxservcnt)
    {
	maxservcnt = servstat;
	maxservtime = time (NULL);
    }
    if (servstat > dailyservstat)
    {
	dailyservstat = servstat;
	dailyservtime = time (NULL);
    }

    /* We keep this in a seperate list because it's easier ;) */
    serverstat = findserv (av[0]);

    if (!serverstat)
    {
	serverstat = scalloc (sizeof (ServStat), 1);

	serverstat->name = sstrdup (av[0]);
	serverstat->version = "Unknown";
	serverstat->flags = "XX";
	serverstat->lastquit = "N/A";
	serverstat->statssince = time (NULL);
	serverstat->lastsplit = time (NULL);
	serverstat->lag.tv_sec = 999;
	serverstat->penalize = 0;
	serverstat->is_uline = finduline (serverstat->name) ? 1 : 0;
	serverstat->is_split = 0;
	serverstat->splittime = 0;
	serverstat->lastdowntime = 0;
	serverstat->splits = 0;
	list = &statlist[SHASH(serverstat->name)];
	serverstat->next = *list;
	if (*list)
	    (*list)->prev = serverstat;
	*list = serverstat;
    }
    else
    {
	/* We'll fill these in just in case this server doesn't reply. */
	serverstat->version = sstrdup ("Unknown");
	serverstat->flags = sstrdup ("XX");

	if (serverstat->is_split)
	{
	    if (serverstat->penalize)
	    {
		serverstat->splittime += time (NULL) - serverstat->lastsplit;
		serverstat->lastdowntime = time (NULL) - serverstat->lastsplit;
		serverstat->penalize = 0;
	    }
	    serverstat->is_split = 0;
	}
    }

    if (finduline (av[0]) && ignoreulines_on == TRUE)
	servstat--;
}

/* Handle an SQUIT command */
static void m_squit (char *source, int ac, char **av)
{
    int i, k, penalize = 1, check = 0;
#if 0
    int j;
    User *u;
#endif
    char tmp[512], *splitserv;
    ServStat *serverstat;

    /* NOQUIT's busted :/ If anyone wants to fix this, go for it..
       the problem is that it doesn't see some users when it goes
       through this while loop, so they don't get freed... If you
       do try to make this work, Be sure to add NOQUIT to the PROTOCTL
       line in main.c (Unreal, DreamForge) or to the CAPAB line in main.c
       (Bahamut, Prometheus). IRCu doesn't support NOQUIT. Also uncomment
       the int and user vars above.
     */     
#if 0
    for (i = 0; i < servcnt && stricmp (servlist[i].server, av[0]); i++);

    if (i < servcnt)
    {
	j = 0;

	/* We give servers the NOQUIT token on link .. If they don't
	   understand it, they'll send QUITs first, so we won't find
	   any users for that server.
	 */
	while (j < 1024)
	{
	    u = userlist[j];

	    if (u && !stricmp (servlist[i].server, u->server))
	    {
		/* If this is a nickname we were waiting for, reclaim it. */
		if (os.wantnick)
		    if (!stricmp (u->nick, os.nick))
		    {
			send_cmd (s_OperServ, "%s %s", me.token ? "&" : "NICK",
			    os.nick);
			strscpy (s_OperServ, os.nick, sizeof (s_OperServ));
			os.wantnick = 0;
		    }

		if (ss.wantnick)
		    if (!stricmp (u->nick, ss.nick))
		    {
			send_cmd (s_StatServ, "%s %s", me.token ? "&" : "NICK",
			    ss.nick);
			strscpy (s_StatServ, ss.nick, sizeof (s_StatServ));
			ss.wantnick = 0;
		    }

		if (gn.wantnick)
		    if (!stricmp (u->nick, gn.nick))
		    {
			send_cmd (s_GlobalNoticer, "%s %s", me.token ? "&" :
			    "NICK", gn.nick);
			strscpy (s_GlobalNoticer, gn.nick, sizeof
			    (s_GlobalNoticer));
			gn.wantnick = 0;
		    }

		delete_user (u);
		u = u->next;
	    }

	    j++;
	}
    }

#if defined (BAHAMUT) || defined (PROMETHEUS)

    /* Bahamut only sends SQUITs for the splitting server, not servers linked
       to it. So we have to free those servers and their users here.
     */
    for (i = 0; i < servcnt; i++)
    {
	if (!stricmp (servlist[i].uplink, av[0]))
	{
	    j = 0;

	    while (j < 1024)
	    {
		u = userlist[j];

		if (u && !stricmp (servlist[i].server, u->server))
		{
		    if (os.wantnick)
			if (!stricmp (u->nick, os.nick))
			{
			    send_cmd (s_OperServ, "%s %s", me.token ? "&" : "NICK",
				os.nick);
			    strscpy (s_OperServ, os.nick, sizeof (s_OperServ));
			    os.wantnick = 0;
			}

		    if (ss.wantnick)
			if (!stricmp (u->nick, ss.nick))
			{
			    send_cmd (s_StatServ, "%s %s", me.token ? "&" : "NICK",
				ss.nick);
			    strscpy (s_StatServ, ss.nick, sizeof (s_StatServ));
			    ss.wantnick = 0;
			}

		    if (gn.wantnick)
			if (!stricmp (u->nick, gn.nick))
			{
			    send_cmd (s_GlobalNoticer, "%s %s", me.token ? "&" :
				"NICK", gn.nick);
			    strscpy (s_GlobalNoticer, gn.nick, sizeof
				(s_GlobalNoticer));
			    gn.wantnick = 0;
			}

		    delete_user (u);
		    u = u->next;
		}

		j++;
	    }

	    /* Find this before we free it. */
	    serverstat = findserv (servlist[i].server);

	    if (debuglevel > 1)
		debug ("m_squit(): Deleting server %s", servlist[i].server);

	    if (servlist[i].server)
		free (servlist[i].server);

	    if (servlist[i].uplink)
		free (servlist[i].uplink);

	    servcnt--;
	    servstat--;

	    if (i < servcnt)
		bcopy (servlist + i + 1, servlist + i, sizeof
		    (*servlist) * (servcnt - i));

	    i--;

	    if (serverstat)
	    {
		serverstat->lastquit = sstrdup (av[1]);
		serverstat->lastsplit = time (NULL);
		serverstat->is_split = 1;
	    }
	}
    }

    /* Free the splitting server */
    for (i = 0; i < servcnt && stricmp (servlist[i].server, av[0]); i++);

    if (debuglevel > 1)
	debug ("m_squit(): Deleting server %s", servlist[i].server);

    if (servlist[i].server)
	free (servlist[i].server);

    if (servlist[i].uplink)
	free (servlist[i].uplink);

    servcnt--;
    servstat--;

    if (i < servcnt)
	bcopy (servlist + i + 1, servlist + i, sizeof
	    (*servlist) * (servcnt - i));

    /* Update the stats for the SQUITting server */
    serverstat = findserv (av[0]);

    if (serverstat)
    {
	/* We know it was the source of the split.. */
	serverstat->splits++;
	serverstat->penalize = 1;
	serverstat->lastquit = sstrdup (av[1]);
	serverstat->lastsplit = time (NULL);
	serverstat->is_split = 1;
    }

    /* We're done, return here */
    return;

#endif /* BAHAMUT */
#endif /* NOQUIT is broken. */

    for (i = 0; i < servcnt && stricmp (servlist[i].server, av[0]); i++);

    if (i < servcnt)
    {
	/* Check the quit message for a period. If there isn't one,
	   We'll assume this is a generic quit message, indicating
	   that the server is at fault (it shut down).
	 */
	if (!strchr (av[1], '.'))
	    penalize = check = 1;

	/* Check if this server is the source of the split, or
	   if it was just lost in a split. splitserv is the source
	   of the split. We're checking here if splitserv was this
	   server's uplink. If it is, then we don't want to penalize
	   this server for being split from the network.
	 */
	if (!check)
	{
	    strscpy (tmp, av[1], sizeof (tmp));
	    strtok (tmp, " ");

	    if (tmp)
		splitserv = strtok (NULL, " ");

	    /* Here's where it gets confusing. I can't think of a reliable
	       way to detect the source of a netsplit, but it's pretty easy
	       to detect which servers AREN'T the source of the split. So,
	       penalize starts out as 1, and if we determine this server isn't
	       the source, then it's 0.
	     */
	    if (splitserv)
	    {
		for (k = 0; k < servcnt; k++)
		    if (stricmp (servlist[i].server, splitserv))
			penalize = 0;
	    }
	}

	if (debuglevel > 1)
	    debug ("m_squit(): Deleting server %s", servlist[i].server);

	if (servlist[i].server)
	    free (servlist[i].server);

	if (servlist[i].uplink)
	    free (servlist[i].uplink);

	servcnt--;
	servstat--;

	if (i < servcnt)
	    bcopy (servlist + i + 1, servlist + i, sizeof
		(*servlist) * (servcnt - i));
    }

    /* We maintain a seperate list for stats */
    serverstat = findserv (av[0]);

    if (serverstat)
    {
	if (penalize)
	{
	    serverstat->splits++;
	    serverstat->penalize = 1;
	}

	serverstat->lastquit = sstrdup (av[1]);
	serverstat->lastsplit = time (NULL);
	serverstat->is_split = 1;
    }
}

/* Handle a KILL command. */
static void m_kill (char *source, int ac, char **av)
{
    if (ac != 2)
	return;

    /* Before we reintroduce the user, we'll check to see if they've
       been collided more than 5 times in the last 5 seconds. This is
       a common problem with people who forget to rename/disable services
       with the same nicks as some of ours. If they HAVE been collided
       that many times, we'll give them a nick like OperServ986662577 and
       send a globops about it. We'll then watch for a nickchange/quit from
       the client with our nick, and retake it if the opportunity presents
       itself.
     */
    if ((operserv_on == TRUE) && (!stricmp (av[0], s_OperServ)))
    {
	/* How many times we've been collided */
	os.collided++;

	/* If this is true, we give up trying to retake our nick.. */
	if (((os.lastcoll >= (time (NULL) - 5)) && (os.collided >= 5)))
	{
	    /* Check if it's a server killing us. If it's not, then it's
	       likely a regular kill, and we don't want to take a waiting
	       nick.
	     */
	    if (!strchr (source, '.'))
	    {
		/* It's not a server. We'll reintro with the regular
		   nick and leave it at that.
		 */
		intro_user (s_OperServ, os.user, os.host, os.real, os.mode);
		return;
	    }

	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       OperServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!os.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (os.nick, s_OperServ, sizeof (os.nick));
	    }

	    /* Make note that we want it */
	    os.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_OperServ, sizeof (s_OperServ), "%s%lu", os.nick,
		time (NULL));

	    /* Introduce with the new nick */
	    intro_user (s_OperServ, os.user, os.host, os.real, os.mode);

	    /* Let the staff know about it */
	    globops (s_OperServ, ERR_COLLISIONS, os.nick);

	    os.collided = 0;

	    return;
	}
	else
	    os.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_OperServ, os.user, os.host, os.real, os.mode);
    }
    else if ((statserv_on == TRUE) && !stricmp (av[0], s_StatServ))
    {
	/* How many times we've been collided */
	ss.collided++;

	/* If this is true, we give up trying to retake our nick.. */
	if (((ss.lastcoll >= (time (NULL) - 5)) && (ss.collided >= 5)))
	{
	    /* Check if it's a server killing us. If it's not, then it's
	       likely a regular kill, and we don't want to take a waiting
	       nick.
	     */
	    if (!strchr (source, '.'))
	    {
		/* It's not a server. We'll reintro with the regular
		   nick and leave it at that.
		 */
		intro_user (s_StatServ, ss.user, ss.host, ss.real, ss.mode);
		return;
	    }

	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       OperServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!ss.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (ss.nick, s_StatServ, sizeof (ss.nick));
	    }

	    /* Make note that we want it */
	    ss.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_StatServ, sizeof (s_StatServ), "%s%lu", ss.nick,
		time (NULL));

	    /* Introduce with the new nick */
	    intro_user (s_StatServ, ss.user, ss.host, ss.real, ss.mode);

	    /* Let the staff know about it */
	    globops (s_StatServ, ERR_COLLISIONS, ss.nick);

	    ss.collided = 0;

	    return;
	}
	else
	    ss.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_StatServ, ss.user, ss.host, ss.real, ss.mode);
    }
    else if ((globalnoticer_on == TRUE) && !stricmp (av[0], s_GlobalNoticer))
    {
	/* How many times we've been collided */
	gn.collided++;

	/* If this is true, we give up trying to retake our nick.. */
	if (((gn.lastcoll >= (time (NULL) - 5)) && (gn.collided >= 5)))
	{
	    /* Check if it's a server killing us. If it's not, then it's
	       likely a regular kill, and we don't want to take a waiting
	       nick.
	     */
	    if (!strchr (source, '.'))
	    {
		/* It's not a server. We'll reintro with the regular
		   nick and leave it at that.
		 */
		intro_user (s_GlobalNoticer, gn.user, gn.host, gn.real,
		    gn.mode);
		return;
	    }

	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       OperServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!gn.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (gn.nick, s_GlobalNoticer, sizeof (gn.nick));
	    }

	    /* Make note that we want it */
	    gn.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_GlobalNoticer, sizeof (s_GlobalNoticer), "%s%lu",
		gn.nick, time (NULL));

	    /* Introduce with the new nick */
	    intro_user (s_GlobalNoticer, gn.user, gn.host, gn.real, gn.mode);

	    /* Let the staff know about it */
	    globops (s_GlobalNoticer, ERR_COLLISIONS, gn.nick);

	    gn.collided = 0;

	    return;
	}
	else
	    gn.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_GlobalNoticer, gn.user, gn.host, gn.real, gn.mode);
    }
    else if ((spamserv_on == TRUE) && !stricmp (av[0], sp.nick))
    {
	char guest[NICKLEN];
	int guestnick;

	/* Unlike above, we don't care if we get collided.. If we do, we'll
	   just pick a new nick. So we don't do any of the fancy collision
	   counting here. We'll make ourselves a new nick and introduce with
         it.
	 */
	guestnick = 1+(int) (99999.0*rand()/(RAND_MAX+10000.0));

	/* Sometimes rand() gives us a 4 digit number. If we get one,
	   we'll add 10000 to it, bringing it into the 5 digit range.
	 */
	if (guestnick < 10000)
	    guestnick = guestnick + 10000;

	snprintf (guest, sizeof (guest), "Guest%d", guestnick);

	/* Now replace the nick */
	strscpy (sp.nick, guest, sizeof (sp.nick));

	/* Finally, introduce the client with the new nick. */
	intro_user (sp.nick, sp.user, sp.host, sp.real, "+");
    }
    else
	do_kill (source, ac, av);
}

/* Handle a nick collision */
static void m_collide (char *source, int ac, char **av)
{
    /* Before we reintroduce the user, we'll check to see if they've
       been collided more than 5 times in the last 5 seconds. This is
       a common problem with people who forget to rename/disable services
       with the same nicks as some of ours. If they HAVE been collided
       that many times, we'll give them a nick like OperServ986662577 and
       send a globops about it. We'll then watch for a nickchange/quit from
       the client with our nick, and retake it if the opportunity presents
       itself.
     */
    if ((operserv_on == TRUE) && (!stricmp (av[1], s_OperServ)))
    {
	/* How many times we've been collided */
	os.collided++;

	/* If this is true, we give up trying to retake our nick.. */
	if (((os.lastcoll >= (time (NULL) - 5)) && (os.collided >= 5)))
	{
	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       OperServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!os.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (os.nick, s_OperServ, sizeof (os.nick));
	    }

	    /* Make note that we want it */
	    os.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_OperServ, sizeof (s_OperServ), "%s%lu", os.nick,
		time (NULL));

	    /* Introduce with the new nick */
	    intro_user (s_OperServ, os.user, os.host, os.real, os.mode);

	    /* Let the staff know about it */
	    globops (s_OperServ, ERR_COLLISIONS, os.nick);

	    os.collided = 0;

	    return;
	}
	else
	    os.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_OperServ, os.user, os.host, os.real, os.mode);
    }
    else if ((statserv_on == TRUE) && !stricmp (av[1], s_StatServ))
    {
	/* How many times we've been collided */
	ss.collided++;

	/* If this is true, we give up trying to retake our nick.. */
	if (((ss.lastcoll >= (time (NULL) - 5)) && (ss.collided >= 5)))
	{
	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       OperServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!ss.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (ss.nick, s_StatServ, sizeof (ss.nick));
	    }

	    /* Make note that we want it */
	    ss.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_StatServ, sizeof (s_StatServ), "%s%lu", ss.nick,
		time (NULL));

	    /* Introduce with the new nick */
	    intro_user (s_StatServ, ss.user, ss.host, ss.real, ss.mode);

	    /* Let the staff know about it */
	    globops (s_StatServ, ERR_COLLISIONS, ss.nick);

	    ss.collided = 0;

	    return;
	}
	else
	    ss.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_StatServ, ss.user, ss.host, ss.real, ss.mode);
    }
    else if ((globalnoticer_on == TRUE) && !stricmp (av[1], s_GlobalNoticer))
    {
	/* How many times we've been collided */
	gn.collided++;

	/* If this is true, we give up trying to retake our nick.. */
	if (((gn.lastcoll >= (time (NULL) - 5)) && (gn.collided >= 5)))
	{
	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       OperServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!gn.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (gn.nick, s_GlobalNoticer, sizeof (gn.nick));
	    }

	    /* Make note that we want it */
	    gn.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_GlobalNoticer, sizeof (s_GlobalNoticer), "%s%lu",
		gn.nick, time (NULL));

	    /* Introduce with the new nick */
	    intro_user (s_GlobalNoticer, gn.user, gn.host, gn.real, gn.mode);

	    /* Let the staff know about it */
	    globops (s_GlobalNoticer, ERR_COLLISIONS, gn.nick);

	    gn.collided = 0;

	    return;
	}
	else
	    gn.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_GlobalNoticer, gn.user, gn.host, gn.real, gn.mode);
    }
    else if ((spamserv_on == TRUE) && !stricmp (av[0], sp.nick))
    {
	char guest[NICKLEN];
	int guestnick;

	/* Unlike above, we don't care if we get collided.. If we do, we'll
	   just pick a new nick. So we don't do any of the fancy collision
	   counting here. We'll make ourselves a new nick and introduce with
         it.
	 */
	guestnick = 1+(int) (99999.0*rand()/(RAND_MAX+10000.0));

	/* Sometimes rand() gives us a 4 digit number. If we get one,
	   we'll add 10000 to it, bringing it into the 5 digit range.
	 */
	if (guestnick < 10000)
	    guestnick = guestnick + 10000;

	snprintf (guest, sizeof (guest), "Guest%d", guestnick);

	/* Now replace the nick */
	strscpy (sp.nick, guest, sizeof (sp.nick));

	/* Finally, introduce the client with the new nick. */
	intro_user (sp.nick, sp.user, sp.host, sp.real, "+");
    }
}

/* Handle a PROTOCTL command */
static void m_protoctl (char *source, int ac, char **av)
{
    int i, token = 0;
#ifdef UNREAL3
    int nickv2 = 0, umode2 = 0;
#endif

    for (i = 0; i < ac; i++)
    {
	if (!stricmp (av[i], "TOKEN"))
	    token = 1;
#ifdef UNREAL3
	if (!stricmp (av[i], "NICKv2"))
	    nickv2 = 1;

	if (!stricmp (av[i], "UMODE2"))
	    umode2 = 1;
#endif
    }

    if (token)
	me.token = 1;

#ifdef UNREAL3
    if (!nickv2 || !umode2)
    {
	send_cmd (NULL,
	    "ERROR :Incompatible IRCd (I'm compiled to support Unreal 3.1+)");
	snprintf (quitmsg, BUFSIZE,
	    "Incompatible IRCd (I'm compiled to support Unreal 3.1+)");
	runflags |= (RUN_QUITTING | RUN_TERMINATING | RUN_SHUTDOWN);
    }
#endif
}

/* Handle a MODE command */
static void m_mode (char *source, int ac, char **av)
{
    if (*av[0] == '#')
    {
	if (ac < 2)
	    return;
	do_cmode (source, ac, av);
    }
    else
    {
	if (ac != 2)
	    return;
	do_umode (source, ac, av);
    }
}

#ifdef UNREAL3
/* Handle an UnrealIRCd UMODE2 command */
static void m_umode2 (char *source, int ac, char **av)
{
    av[1] = sstrdup (av[0]);
    strcpy (av[0], source);

    do_umode (source, ac, av);
}
#endif

#ifdef PROMETHEUS
/* Handle the ENDBURST command */
static void m_endburst (char *source, int ac, char **av)
{
   globops (operserv_on == TRUE ? s_OperServ : s_StatServ,
	"Sync'd to network in %s: %d users, %d channels, %d servers.",
	duration (time (NULL) - burststart, 1), userstat, chanstat,
	servstat);
}
#endif

/* Handle PRIVMSGs. */
static void m_privmsg (char *source, int ac, char **av)
{
    char buf[BUFSIZE];

    if (ac != 2)
	return;

    snprintf (buf, sizeof (buf), "%s@%s", s_OperServ, me.name);
    if (!stricmp (av[0], s_OperServ) || !stricmp (av[0], buf))
    {
	operserv (source, av[1]);
	return;
    }

    snprintf (buf, sizeof (buf), "%s@%s", s_StatServ, me.name);
    if (!stricmp (av[0], s_StatServ) || !stricmp (av[0], buf))
    {
	statserv (source, av[1]);
	return;
    }

    snprintf (buf, sizeof (buf), "%s@%s", s_GlobalNoticer, me.name);
    if (!stricmp (av[0], s_GlobalNoticer) || !stricmp (av[0], buf))
    {
	globalnoticer (source, av[1]);
	return;
    }

    snprintf (buf, sizeof (buf), "%s@%s", sp.nick, me.name);
    if (!stricmp (av[0], sp.nick) || !stricmp (av[0], buf))
    {
	spamserv (source, av[1]);
	return;
    }
}

Message messages[] = {
    {"436",	m_collide},
    {"KICK",	m_kick},
    {"KILL",	m_kill},
    {"MODE",	m_mode},
    {"NICK",	m_nick},
    {"PART",	m_part},
    {"PING",	m_ping},
    {"PONG",	m_pong},
    {"PRIVMSG",	m_privmsg},
    {"JOIN",	m_join},
    {"QUIT",	m_quit},
    {"SERVER",	m_server},
    {"SQUIT",	m_squit},
    {"VERSION",	m_version},
    {"INFO",	m_info},
    {"WHOIS",	m_whois},
    {"STATS",	m_stats},
    {"MOTD",	m_motd},
    {"TIME",	m_time},
    {"351",	m_vreply},
    {"ADMIN",	m_admin},
    {"AKILL",	m_akill},
    {"RAKILL",	m_rakill},
    {"TOPIC",	m_topic},
    {"PROTOCTL",m_protoctl},

    /* IRCd specific commands */
#ifdef UNREAL3
    {"NETINFO",	m_netinfo},
    {"UMODE2",	m_umode2},
#endif

#if defined (BAHAMUT) || defined (PROMETHEUS)
    {"SJOIN",	m_sjoin},
#endif

#ifdef PROMETHEUS
    {"EB",	m_endburst},
#endif

    /* Tokens - DreamForge - DreamForge has an array of tokens defined,
       but only ever sends 5 tokenized commands, one of which we don't
       care about.
     */
    {"G",	m_mode},
    {"&",	m_nick},
    {"C",	m_join},
    {")",	m_topic},

    /* Tokens - UnrealIRCd. - Unreal sends more than just these, but 
       these are the only ones we care about right now. Others will be
       added later as we add new features. The 4 above are the same on
       Unreal.
     */
#ifdef UNREAL3
    {"H",       m_kick},
    {".",	m_kill},
    {"+",	m_version},
    {"/",	m_info},
    {"-",	m_squit},
    {"#",	m_whois},
    {"2",	m_stats},
    {">",	m_time},
    {"@",	m_admin},
    {"F",	m_motd},
    {"D",	m_part},
    {"8",	m_ping},
    {"9",	m_pong},
    {",",	m_quit},
    {"'",	m_server},
    {"!",	m_privmsg},
    {"|",	m_umode2},
    {"AO",	m_netinfo},
#endif /* UNREAL3 */    
    {NULL}
};

Message *find_message (const char *name)
{
    Message *m;

    for (m = messages; m->name; m++)
	if (!stricmp (name, m->name))
	    return m;

    return NULL;
}
